/***************************************************************************
 *
 * SortedDoublyLinkedList.java
 *
 * More advanced implementation of the List interface using a doubly
 * linked list with references to both the head and tail Node's in the list
 *
 *****************************************************************************/

import java.util.*;


public class SortedDoublyLinkedList<E extends Comparable<? super E>> implements Iterable<E> {
// Remark: All Implemented Interfaces:
//java.lang.Iterable<E>, java.util.Collection<E>, java.util.List<E> 

 /*******************************************************
 *
 *  The Node class
 *
 ********************************************************/
   protected static class Node<E> {
      private E data;
      private Node<E> next;
      private Node<E> prev;

      public Node(E data, Node<E> next, Node<E> prev) {
         this.data = data;
         this.next = next;
         this.prev = prev;
      }
   }  //end-class-Node
   
  /*******************************************************
  *
  *  The Iterator class
  *
  ********************************************************/

     public Iterator<E> iterator()
     {
        return new LinkedListIterator();
     }


     private class LinkedListIterator  implements Iterator<E>
     {
        private Node<E> nextNode;

        public LinkedListIterator()
        {
           nextNode = head;
        }

        public boolean hasNext()
        {
           return nextNode != null;
        }

        public E next()
        {
           if (!hasNext()) throw new NoSuchElementException();
           E res = nextNode.data;
           nextNode = nextNode.next;
           return res;
        }

        public void remove() { throw new UnsupportedOperationException(); }

     }  //end class LinkedListIterator



     
   private Node<E> head = null;
   private Node<E> tail = null;
   private int size = 0;
   private Comparator<E> c = null;

   /**
    *  Constructs an empty list
    */
    public SortedDoublyLinkedList() { }
    
    /**
     *  Constructs an empty list
     */
    public SortedDoublyLinkedList(Comparator<E> comparator)
    {
      c = comparator;
    }

    
  /**
   *  Returns true if the list is empty
   */
   public boolean isEmpty()  {  return size == 0; }

  /**
   *  Returns the length of the list
   */
   public int size()  { return size; }
   
  /**
   *  Gets the value of the property comparator
   */
   Comparator<E> getComparator()  { return c; }
    
    
    /**
     *  Sets the value of the property comparator
     */
      public void  setComparator(Comparator<E> comparator)
     {
        	SortedDoublyLinkedList<E> list = insertionSort(comparator);
        	c = comparator;
        	head = list.head;
        	tail = list.tail;
     }
      
      
     /** Inserts the element into the proper sorted location.
      */
     public void insert(E item)
     { 
    	 insert(item,c);
     
     }
     

    private boolean lessThan(E x, E y, Comparator<E> comparator){
   	if (comparator == null) return x.compareTo(y) < 0;
   	else return comparator.compare(x, y) < 0;
     }
     
    
     public void insert(E item, Comparator<E> comparator) { 
    	  // se comparator e' null si considera l'ordinamento naturale
  	    
    	    if (size == 0) {
    	    	tail = head = new Node<E>(item, null, null);
    	    	size++;
    	    	return;
    	    }
    	    if (lessThan(item, head.data, comparator)) {
      	      head = new Node<E>(item, head, null);
      	      head.next.prev = head;
  	    	  size++;
  	    	  return;
      	    }
    	   
    	    Node<E> cur = head;
    	    while (cur!= null && lessThan(cur.data, item, comparator)) cur = cur.next;
    	    if (cur == null) {
    	    	   // insert at the last position
    	    	   tail.next = new Node<E>(item, null, tail);
    	    	   tail = tail.next;
    	    } else { // item precedes cur node 
        		   cur.prev = new Node<E>(item, cur, cur.prev);
        		   cur.prev.prev.next = cur.prev;
    	    }

    	    size ++;
    	  } //end-insert



     /**
      *  A Comparator based ordering method
      */
     
     public SortedDoublyLinkedList<E> insertionSort()
    // se comparator e' null si considera l'ordinamento naturale
     {
   	   return insertionSort(null);
     }
     
     
      public SortedDoublyLinkedList<E> insertionSort(Comparator<E> comparator)
     // se comparator e' null si considera l'ordinamento naturale
      {
         SortedDoublyLinkedList<E> list = new SortedDoublyLinkedList<E>();
         for (E item: this)  list.insert( item, comparator);
         return list;
      }      
      
      

 
 /**
   *  Returns the first element in the list.
   *
   */
   public E getFirst()
   {
      if(head == null) throw new NoSuchElementException();
      return head.data;
   }

 /**
   *  Removes the first element in the list.
   *
   */
   public E removeFirst()
   {
	  if (head == null) throw new NoSuchElementException();
      E elem = head.data;
      head = head.next;
      if (head == null) tail = null;
      else head.prev = null;
      size--;
      return elem;
   }

  /**
   *  Returns the last element in the list.
   *
   */
   public E getLast()
   {
      if(tail == null) throw new NoSuchElementException();
      return tail.data;
   }


 /**
   *  Removes the last element in the list.
   *
   */
   public E removeLast()
   {  
    if(tail == null) throw new NoSuchElementException();
	E elem = tail.data;
    if (size == 1) return removeFirst();
	// there are at least two nodes
    tail = tail.prev;
    tail.next = null;
    size--;
    return elem;
   }


 /**
   *  Removes all nodes from the list.
   *
   */
   public void clear()
   {
      head = tail = null; size=0;
   }


 /**
   *  Returns the data at the specified position in the list.
   *
   */
   public E get(int index)
   {
      if (head == null || index > size ) throw new IndexOutOfBoundsException();
      Node<E> cur = head;
      for (int k = 0; k < index; k++) cur = cur.next;
      return cur.data;
   }


 /**
   *  Returns a string representation
   *
   */
   public String toString()
   {
      StringBuffer result = new StringBuffer();  
      for(Object x : this)
      	result.append(x + "; ");

      return result.toString();
   }
      

 /**
   *  Removes the first occurrence of the specified element in this list.
   *
   */
   public boolean remove(Object key)
   {
      return false;	
   }
   
   
   public boolean remove(E key)
   {
    if ( (head == null) || lessThan(key, head.data, this.c) ) return false;

    Node<E> cur  = head;
    while(cur != null && lessThan(cur.data, key, this.c)) cur = cur.next;
    if(cur == null || !cur.data.equals(key)) return false;
      
      //delete cur node
    if (cur == head) removeFirst();
    else if (cur == tail) removeLast();
    else {
    	  cur.prev.next = cur.next;
    	  cur.next.prev = cur.prev;
    }
    size--;
    return true;	
   }
   
   /**
    *  Removes the first occurrence of the specified element in this list.
    *
    */
    public boolean removeRec (E key)
    {
       if ( (head == null) || lessThan(key, head.data, this.c) ) return false;
       if (head.data.equals(key)) { 
     	  removeFirst();
     	  return true; 
       } else return removeRec(head.next, key);
    }


    private boolean removeRec (Node<E> node, E key){
       if ( (node == null) ||  lessThan(key, node.data, this.c) ) return false;
       if (node.data.equals(key)) {
    	    if (node == tail) removeLast();
    	    else {
    	    	  node.prev.next = node.next;
    	    	  node.next.prev = node.prev;
    	    }
     	  size--;
     	  return true;
       }
       return removeRec(node.next, key);
        	
    }
    
    
    
    /**
     *  Removes the element at the specified position in this list
     */   

       public E remove (int index) {
          if (head == null || index > size ) throw new IndexOutOfBoundsException(); 

          if (index == 0) return this.removeFirst();
          System.out.println("\nGiovanna");
          
          Node<E> cur = head.next; // index 1          
          for (int k = 0; k < index; k++) cur = cur.next;
          //delete current node
    	    E obj = cur.data;
    	   	if (cur == tail) {
    	   		tail = cur.prev;
    	   		tail.next = null;
    	   	} else
    	   	{
              cur.prev.next = cur.next;
              cur.next.prev = cur.prev;
    	   	}
              size--;
              return obj;
       }
       
       
    
    /**
     *  Returns true if this list contains the specified element.
     *
     */
    public boolean contains(Object x) {return false;}
    
    public boolean contains(E x)  
     {
        for(E curEl : this)
          if(curEl.equals(x)) return true;
          else if (lessThan(x, curEl, this.c)) return false;

        return false;
     }
    

     /**
       *  Replaces the element at the specified position in this list
       *  with the specified element.
       *  Returns the element previously at the specified position
       */
       public E set(int index, E element)
       { throw new UnsupportedOperationException(); }       
                

         /**
           *  Returns a copy of the immutable list
           *  Complexity: O(n)
           */
           public  SortedDoublyLinkedList<E> copy()
           {
              SortedDoublyLinkedList<E> twin = new SortedDoublyLinkedList<E>();
              if (size == 0) return twin;
              Iterator<E> itr = this.iterator();
              // add first node
              twin.tail = twin.head = new Node<E>(itr.next(), null, null);
              while (itr.hasNext()) {
            	  // addLast
            	  twin.tail.next = new Node<E>(itr.next(), null, twin.tail);
            	  twin.tail = twin.tail.next; 
              }
              return twin;
           }

  /**
   * toArray() method returns an array containing all of
   * the elements in this list in proper sequence (from
   * first to last element). 
   * */ 

   public Object[] toArray() {
         Object[] array;
         int i=0;	
         if (head == null) return null;
         array = new Object[size];
         for (E item: this) {array[i++] = item; }
         return array;

     }
   
   
   public boolean equals (Object obj) {
	   if (obj == null) return false;
	   if (! (obj instanceof SortedDoublyLinkedList<?>)) return false;
	   SortedDoublyLinkedList<?> dlist = (SortedDoublyLinkedList<?>)obj;
	   Iterator<E> itr1 =  this.iterator();
	   Iterator<?> itr2 = dlist.iterator();
	   while (itr1.hasNext() && itr2.hasNext())
		   if ( !itr1.next().equals(itr2.next()) ) return false;
	   if (itr1.hasNext() || itr2.hasNext()) return false;
	   else return true;
	   
   }
   

  
        /*
        *   Unsupported methods
        */
          public List<E> subList(int fromIndex, int toIndex)
          // Returns a view of the portion of this list between the specified fromIndex, inclusive, and toIndex, exclusive. 
          { throw new UnsupportedOperationException(); }

          public int indexOf(Object o)
          // Returns the index of the first occurrence of the specified element in this list,
          // or -1 if this list does not contain the element.
          { throw new UnsupportedOperationException(); }

          public int lastIndexOf(Object o)
          // Returns the index of the last occurrence of the specified element in this list,
          // or -1 if this list does not contain the element.
          { throw new UnsupportedOperationException(); }

          public ListIterator<E> listIterator()
            	// Returns a list iterator over the elements in this list (in proper sequence)
          { throw new UnsupportedOperationException(); }

          public ListIterator<E> listIterator(int index)
            	// Returns a list iterator over the elements in this list (in proper sequence),
       	// starting at the specified position in the list.
          { throw new UnsupportedOperationException(); }

          public boolean retainAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public boolean removeAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public boolean addAll(Collection<? extends E> c)
          { throw new UnsupportedOperationException(); }

          public boolean addAll(int index, Collection<? extends E> c)
          { throw new UnsupportedOperationException(); }

          public boolean containsAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public <T> T[] toArray(T[] a)
          { throw new UnsupportedOperationException(); }

} // end-class 